// code by jph

#ifndef MESH_INCLUDED
#define MESH_INCLUDED

#define MESH_VERT      0      //indexes the ind[.] array which stores vertex indices for each shape
#define MESH_EDGE      1
#define MESH_TRIA      2
#define MESH_QUAD      3
#define MESH_TETR      4
#define MESH_OCTA      5
#define MESH_PRIS      6
#define MESH_CUBE      7
#define MESH_MAX       8

#define MESH_VERT_     1      //the same again for bitwise occupation, as used in drawing routine
#define MESH_EDGE_     2                                             
#define MESH_TRIA_     4
#define MESH_QUAD_     8
#define MESH_TETR_     16
#define MESH_OCTA_     32
#define MESH_PRIS_     64
#define MESH_CUBE_     128
#define MESH_ALL       255

static int MESH_OCCUP  []={1,2,4,8,16,32,64,128};  //bitwise access in for loops

static char* MESH_INFO  []={"VERT","EDGE","TRIA","QUAD","TETR","OCTA","PRIS","CUBE"};

#define MESH_ZIPPER    1      //for example subdivide(MESH_ZIPPER|MESH_LINEAR|MESH_AVERAGE)
#define MESH_LINEAR    2      //not all combinations are possible, I forgot which,
#define MESH_AVERAGE   4      //but subdivide(MESH_LINEAR) works.
#define MESH_NORMAL    7

#define MESH_FORCE     256    //attribute for PHANToM operation
#define MESH_JAYA      1024    //file input from jaya mesh
#define MESH_TXT       2048    //file input from text format mesh

#define MESH_NORMALS   256    //has currently no effect
#define MESH_VERTICES  512
#define MESH_TOPOLOGY  1024
#define MESH_NOCOLOR   1024
#define MESH_NODEPTH   2048    //disables depth buffer in mesh drawing routine
#define MESH_OPAQUE    8192    //displays the shapes not transparent
#define MESH_LABEL     16384  //labels vertices with their index
#define MESH_EDGES     32768  //uses also next 2 bits to encode size
#define MESH_GIANT     536870912 //alternatively 67108864

static int    MESH_D       []={0,3,12,240,MESH_ALL};  //used for addressing all of dim i
static int    MESH_COMP    []={1,2,3,4,4,6,6,8};      //number of vertices at shape
static int    MESH_ADD_VPF []={0,0,1,2,2,3,4,7};      //number of vertices essencially added by face REVISE!!!
static int    MESH_REPRO   []={1,2,4,4,4,6,8,8};      //number of face type after lin sub
static int    MESH_REPOSIT []={1,3,6,9,10,19,18,27};  //vertices in face at lin sub
static int    MESH_DIMS    []={0,1,2,2,3,3,3,3};      //dimensionality of shapes
static int    MESH_JAYAIND []={9,MESH_VERT,MESH_EDGE,MESH_TRIA,MESH_TETR,9,MESH_OCTA,9}; //the ones with nine dont matter
static int    MESH_CLASS   []={MESH_VERT,MESH_EDGE,MESH_TRIA,MESH_QUAD,MESH_TETR,MESH_TETR,MESH_PRIS,MESH_CUBE}; //vertex classes
static double MESH_AVERAGEC[]={        1,       .5,1.0/6.0,     .25,1.0/14.0, 1.0/14.0,1.0/12.0,1.0/8.0};

static char   MESH_BUFFER[1024];

#define       MESH_DEFAULT      format=0; \
                                texture=0; \
                                color=tensor<double>(1,4).as( 1.0, 1.0, 1.0, 1.0); \
                                specular=tensor<double>(1,4).as( 1.0, 1.0, 1.0, 1.0); \
                                shininess=20; // a value between 1 and 128 (128 being the brightest intensity)

template<class type>
class mesh {
public:
  mesh<type>();
  mesh<type>(const mesh<type>& myMesh);
  mesh<type>(char* name);                 //retrieves mesh from binary files fn.extentions
  void store(char* name);
  const mesh<type>& operator=(const mesh<type>& myMesh);
  char* toInfo(void);

  tensor<int>     ind[MESH_MAX];        // topology
  tensor<type>    arr;                  // vertex array
  static int      valid;

  uint            format;               // optional (!i.e. GL_T2F_C4F_N3F_V3F!)
  uint            texture;              // optional
  tensor<double>  color;
  tensor<double>  specular;
  double          shininess;
};

//--------------------------------------------------------------------------------------
template<class type>
mesh<type>::mesh(void) {
  for (int c0=0;c0<MESH_MAX;++c0)
    ind[c0]=tensor<int>(2,MESH_COMP[c0],0);
  arr=tensor<type>(2,0,0);
  MESH_DEFAULT
}

template<class type>
mesh<type>::mesh(const mesh<type>& myMesh) {
  for (int c0=0;c0<MESH_MAX;++c0)
    ind[c0]=myMesh.ind[c0];
  arr=myMesh.arr;
  format=myMesh.format;
  texture=myMesh.texture;
  color=myMesh.color;
  specular=myMesh.specular;
  shininess=myMesh.shininess;
}

template<class type>
mesh<type>::mesh(char *name) {
  TENSOR_REPORT=0;
  sprintf(MESH_BUFFER,"%s.arr",name);
  arr=tensor<type>(MESH_BUFFER);
  valid=!!arr.data;
  if (!valid)
    arr=tensor<type>(2,0,0);
    //printf("ERROR %s\n",fn);  
  for (int c0=0;c0<MESH_MAX;++c0) {
    sprintf(MESH_BUFFER,"%s_%i.ind",name,c0);
    ind[c0]=tensor<int>(MESH_BUFFER);
    if (!ind[c0].data)
      ind[c0]=tensor<int>(2,MESH_COMP[c0],0);
    else
      for (int c1=0;c1<ind[c0].numel;++c1)
        assert(ind[c0].data[c1]<arr.numel);
  }
  TENSOR_REPORT=1;
  MESH_DEFAULT
}

template<class type>
void mesh<type>::store(char* name) {
  sprintf(MESH_BUFFER,"%s.arr",name);
  arr.store(MESH_BUFFER);
  for (int c0=0;c0<MESH_MAX;++c0) {
    sprintf(MESH_BUFFER,"%s_%i.ind",name,c0);
    if (ind[c0].numel)
      ind[c0].store(MESH_BUFFER);
    else
      remove(MESH_BUFFER);
  }
}

template<class type>
const mesh<type>& mesh<type>::operator=(const mesh<type>& myMesh) {
  if (this==&myMesh)
    return *this;
  for (int c0=0;c0<MESH_MAX;++c0)
    ind[c0]=myMesh.ind[c0];
  arr=myMesh.arr;
  format=myMesh.format;
  texture=myMesh.texture;
  return *this;
}

template<class type>
char* mesh<type>::toInfo(void) {
  int len=sprintf(MESH_BUFFER,"mesh<%i>(\n",sizeof(type));
  for (int c0=0;c0<MESH_MAX;c0++)
    if (ind[c0].numel)
      len+=sprintf(MESH_BUFFER+len," %s %s\n",MESH_INFO[c0],ind[c0].toInfo());
  len+=sprintf(MESH_BUFFER+len," %s\n)",arr.toInfo());
  return MESH_BUFFER;
}

int mesh<float>::valid=1;
int mesh<double>::valid=1;

#endif
