// code by jph

// * default formats encouraged: GL_T2F_N3F_V3F, or GL_T2F_C4F_N3F_V3F

class meshgl : public mesh<float> {
public:
  meshgl();
  meshgl(char *fn);
  
  static meshgl surface(const tensor<float>& val);
  
  int offsetT2F(void);
  int offsetC3F(void);
  int offsetN3F(void);
  int offsetV3F(void);
  
  meshgl map(tensor<double> mat); // transforms N3F and V3F (!does not normalize N3F!)
  void scale(float fac);
  void setC3F(tensor<double> col);
  void setC4F(tensor<double> col);
  void updateN3F(void);
  void draw(void);
  void drawN3F(float sca,int skip);

  tensor<double> centroid(double& vol); // vector(4)  with  x y z 1, and volume
  tensor<double> inertia(void);
  tensor<double> bounding(void);
  tensor<double> vertices(void);
  char* toInfo(void);
  
  // -------------------------------------- meshgl_edit.hpp
  //void editable(void);
  //int  addVertex(tensor<double> pnt);
  void remVertex(int val);
  tensor<int> references(void);
  meshgl reduce(void);
  meshgl collapse(double thr);                 // very slow !!!
  meshgl collapse(tensor<int> ref,double thr); // very slow !!!
  meshgl mirror(tensor<int> idx);
  meshgl cut(void); // cuts surface mesh along plane z==0; does not reduce() the mesh returned

  // only for quick evaluation:
  tensor<double> point(int idx);
  tensor<double> vector(int idx);
  tensor<double> normal(int idx);
};

meshgl::meshgl() : mesh<float>() {}

meshgl::meshgl(char* fn) : mesh<float>(fn) {
  //assert(mesh<float>::valid);
  if (mesh<float>::valid && arr.numel)
    switch (arr.size[0]) {
      case  3: format=GL_V3F;             break;
      //case  4: format=GL_V4F;             break;
      case  5: format=GL_T2F_V3F;         break;
      case  6: format=GL_N3F_V3F;         break; // ambigous with C3F_V3F
      case  8: format=GL_T2F_N3F_V3F;     break; // GL_T2F_C3F_V3F is ignored
      case 10: format=GL_C4F_N3F_V3F;     break;
      case 12: format=GL_T2F_C4F_N3F_V3F; break;
      case 15: format=GL_T4F_C4F_N3F_V4F; break;
      default:
        assert(false);
    }
}

// TODO make C4F optional
meshgl meshgl::surface(const tensor<float>& val) {
  int c0,c1,c2=0,c3=0;
  meshgl ret;
  ret.ind[MESH_QUAD]=tensor<int>(2,MESH_COMP[MESH_QUAD],(val.size[0]-1)*(val.size[1]-1));
  ret.arr=tensor<float>(2,8,val.numel);
  ret.format=GL_T2F_N3F_V3F;
  float tdx=1/(float)(val.size[0]-1);
  float tdy=1/(float)(val.size[1]-1);
  for (c1=0;c1<val.size[1];++c1)
  for (c0=0;c0<val.size[0];++c0) {
    ret.arr(0,c2)=c0*tdx;    // texture
    ret.arr(1,c2)=c1*tdy;    // texture
    // normals assigned via updateN3F();
    ret.arr(5,c2)=c0;                           // vertices
    ret.arr(6,c2)=c1;                           // vertices
    ret.arr(7,c2)=val.data[c0+c1*val.size[0]];  // vertices
    if (c0&&c1)
      ret.ind[MESH_QUAD](0,c3++)=c2;            // used in the next loop
    ++c2;
  }
  assert(ret.ind[MESH_QUAD].size[1]==c3);
  for (c0=0;c0<ret.ind[MESH_QUAD].size[1];++c0) {
    c1=ret.ind[MESH_QUAD](0,c0);
    ret.ind[MESH_QUAD](1,c0)=c1-1;
    ret.ind[MESH_QUAD](2,c0)=c1-val.size[0]-1;
    ret.ind[MESH_QUAD](3,c0)=c1-val.size[0];
  }
  for (c0=0;c0<ret.ind[MESH_QUAD].numel;++c0)
    assert(0<=ret.ind[MESH_QUAD].data[c0]&&ret.ind[MESH_QUAD].data[c0]<ret.arr.size[1]);
  ret.updateN3F();
  return ret;
}

int meshgl::offsetT2F(void) {
  switch (format) {
    case GL_T2F_V3F:          return 0;
    case GL_T2F_C4UB_V3F:     return 0;
    case GL_T2F_C3F_V3F:      return 0;
    case GL_T2F_N3F_V3F:      return 0;
    case GL_T2F_C4F_N3F_V3F:  return 0;
    case GL_T4F_C4F_N3F_V4F:  return 0;
  }
  assert(false);
  return 0;
}

int meshgl::offsetC3F(void) {
  switch (format) {
    case GL_C3F_V3F:          return 0;
    case GL_C4F_N3F_V3F:      return 0;
    case GL_T2F_C3F_V3F:      return 2;
    case GL_T2F_C4F_N3F_V3F:  return 2;
    case GL_T4F_C4F_N3F_V4F:  return 4;
  }
  assert(false);
  return 0;
}

int meshgl::offsetN3F(void) {
  switch (format) {
    case GL_N3F_V3F:          return 0;
    case GL_C4F_N3F_V3F:      return 4;
    case GL_T2F_N3F_V3F:      return 2;
    case GL_T2F_C4F_N3F_V3F:  return 6;
    case GL_T4F_C4F_N3F_V4F:  return 8;
  }
  assert(false);
  return 0;
}

int meshgl::offsetV3F(void) {
  switch (format) {
    case GL_V3F:              return 0;
    //case GL_V4F:              return 0;
    case GL_C3F_V3F:          return 3;
    case GL_N3F_V3F:          return 3;
    case GL_C4F_N3F_V3F:      return 7;
    case GL_T2F_V3F:          return 2;
    case GL_T2F_C4UB_V3F:     return 3;
    case GL_T2F_C3F_V3F:      return 5;
    case GL_T2F_N3F_V3F:      return 5;
    case GL_T2F_C4F_N3F_V3F:  return 9;
    case GL_T4F_C4F_N3F_V4F:  return 11;
  }
  printf("format error %d\n",format);
  assert(false);
  return 0;
}

// applies 4x4 affine transformation mat to strips: N3F V3F
meshgl meshgl::map(tensor<double> mat) {
  meshgl ret(*this);
  int n3f=offsetN3F();
  int v3f=offsetV3F();
  tensor<float> vec(1,4); // for n3f
  tensor<float> pnt(1,4); pnt(3)=1; // for v3f
  for (int c0=0;c0<arr.numel;c0+=arr.size[0]) {
    vec.in(0,&arr.data[c0+n3f],3);
    ret.arr.in(c0+n3f,(mat*vec.to<double>()).to<float>().data,3);    
    pnt.in(0,&arr.data[c0+v3f],3);
    ret.arr.in(c0+v3f,(mat*pnt.to<double>()).to<float>().data,3);
  }
  return ret;
}

void meshgl::scale(float fac) {
  for (int c0=offsetV3F();c0<arr.numel;c0+=arr.size[0]) {
    arr.data[c0  ]*=fac;
    arr.data[c0+1]*=fac;
    arr.data[c0+2]*=fac;
  }
}

void meshgl::setC3F(tensor<double> col) {
  tensor<float> cof=col.to<float>();
  for (int c0=offsetC3F();c0<arr.numel;c0+=arr.size[0])
    arr.in(c0,cof.data,3);
}

void meshgl::setC4F(tensor<double> col) {
  tensor<float> cof=col.to<float>();
  for (int c0=offsetC3F();c0<arr.numel;c0+=arr.size[0])
    arr.in(c0,cof.data,4);
}

#define MESHGL_UPDATE_N3F \
    v10=arr.data[v3f+t1  ]-arr.data[v3f+t0  ]; \
    v11=arr.data[v3f+t1+1]-arr.data[v3f+t0+1]; \
    v12=arr.data[v3f+t1+2]-arr.data[v3f+t0+2]; \
    v20=arr.data[v3f+t2  ]-arr.data[v3f+t0  ]; \
    v21=arr.data[v3f+t2+1]-arr.data[v3f+t0+1]; \
    v22=arr.data[v3f+t2+2]-arr.data[v3f+t0+2]; \
    v00=v11*v22-v12*v21; \
    v01=v12*v20-v10*v22; \
    v02=v10*v21-v11*v20; \
    arr.data[t0+=n3f]+=v00; \
    arr.data[++t0   ]+=v01; \
    arr.data[++t0   ]+=v02; \
    arr.data[t1+=n3f]+=v00; \
    arr.data[++t1   ]+=v01; \
    arr.data[++t1   ]+=v02; \
    arr.data[t2+=n3f]+=v00; \
    arr.data[++t2   ]+=v01; \
    arr.data[++t2   ]+=v02;

void meshgl::updateN3F(void) {
  int t0,t1,t2;
  int n3f=offsetN3F();
  int v3f=offsetV3F();
  for (int c0=n3f;c0<arr.numel;c0+=arr.size[0])
    arr.data[c0]=arr.data[c0+1]=arr.data[c0+2]=0;
  float v00,v01,v02;
  float v10,v11,v12;
  float v20,v21,v22;
  int sz0=arr.size[0];
  for (int c0=0;c0<ind[MESH_TRIA].numel;c0+=3) {
    t0=ind[MESH_TRIA].data[  c0]*sz0;
    t1=ind[MESH_TRIA].data[1+c0]*sz0;
    t2=ind[MESH_TRIA].data[2+c0]*sz0;
    MESHGL_UPDATE_N3F
  }
  for (int c0=0;c0<ind[MESH_QUAD].numel;c0+=4) {
    t0=ind[MESH_QUAD].data[3+c0]*sz0;
    t1=ind[MESH_QUAD].data[  c0]*sz0;
    t2=ind[MESH_QUAD].data[1+c0]*sz0;
    MESHGL_UPDATE_N3F
  }
  for (int c0=1;c0<ind[MESH_QUAD].numel;c0+=4) {
    t0=ind[MESH_QUAD].data[  c0]*sz0;
    t1=ind[MESH_QUAD].data[1+c0]*sz0;
    t2=ind[MESH_QUAD].data[2+c0]*sz0;
    MESHGL_UPDATE_N3F
  }
  float nrm;
  for (int c0=n3f;c0<arr.numel;c0+=arr.size[0]) {
    nrm=sqrt(arr.data[c0]*arr.data[c0]+arr.data[c0+1]*arr.data[c0+1]+arr.data[c0+2]*arr.data[c0+2]);
    nrm=nrm<TENSOR_NO_DIVISION?1:1/nrm;
    arr.data[c0]*=nrm;
    arr.data[c0+1]*=nrm;
    arr.data[c0+2]*=nrm;
  }
}

void meshgl::draw(void) {
  // http://www.opengl.org/discussion_boards/ubbthreads.php?ubb=showflat&Number=261112
  glMaterialfv(GL_FRONT,GL_SPECULAR,specular.to<float>().data);
  glMaterialf(GL_FRONT,GL_SHININESS,shininess);
    glEnable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D,texture);
  glColor4dv(color.data);
  glInterleavedArrays(format,arr.size[0]*4,arr.data);
  glDrawElements(GL_POINTS    ,ind[MESH_VERT].numel,GL_UNSIGNED_INT,ind[MESH_VERT].data);
  glDrawElements(GL_LINES     ,ind[MESH_EDGE].numel,GL_UNSIGNED_INT,ind[MESH_EDGE].data);
  glDrawElements(GL_TRIANGLES ,ind[MESH_TRIA].numel,GL_UNSIGNED_INT,ind[MESH_TRIA].data);
  glDrawElements(GL_QUADS     ,ind[MESH_QUAD].numel,GL_UNSIGNED_INT,ind[MESH_QUAD].data);    
    glDisable(GL_TEXTURE_2D);
}

void meshgl::drawN3F(float sca,int skip) {
  skip=arr.size[0]*++skip;
  //int c3f=offsetC3F();
  int n3f=offsetN3F();
  int v3f=offsetV3F();
  tensor<float> pnt(1,3);
  tensor<float> nrm(1,3);
  glBegin(GL_LINES);
    for (int c0=0;c0<arr.numel;c0+=skip) {
      //glColor3fv(&arr.data[c0+c3f]);
      nrm.in(0,&arr.data[c0+n3f],3);
      glNormal3fv(nrm.data);
      glVertex3fv(pnt.in(0,&arr.data[c0+v3f],3).data);
      glVertex3fv((pnt+nrm*sca).data);
    }
  glEnd();
}

#define MESHGL_CENTROID   \
    trd=trh.to<double>();             \
    det=trd.det();                    \
    vol+=det;                         \
    add.in(0,trd.mean().data,3);      \
    ret+=add*det;


tensor<double> meshgl::centroid(double& vol) {
  tensor<int> idx=tensor<int>::index(1,ind[MESH_QUAD].size[1]);
  int v3f=offsetV3F();
  tensor<float> trh(2,3,3);
  tensor<double> trd;
  tensor<double> ret(1,4);
  tensor<double> add(1,4); add(3)=4/3.0;
  vol=0;
  double det;
  // TRIA
  for (int c0=0;c0<ind[MESH_TRIA].numel;c0+=3) {
    trh.in(0,&arr(v3f,ind[MESH_TRIA].data[0+c0]),3);
    trh.in(3,&arr(v3f,ind[MESH_TRIA].data[1+c0]),3);
    trh.in(6,&arr(v3f,ind[MESH_TRIA].data[2+c0]),3);
    MESHGL_CENTROID
  }
  // QUAD
  for (int c0=0;c0<ind[MESH_QUAD].numel;c0+=4) {
    trh.in(0,&arr(v3f,ind[MESH_QUAD].data[3+c0]),3);
    trh.in(3,&arr(v3f,ind[MESH_QUAD].data[0+c0]),3);
    trh.in(6,&arr(v3f,ind[MESH_QUAD].data[1+c0]),3);
    MESHGL_CENTROID
    trh.in(0,&arr(v3f,ind[MESH_QUAD].data[1+c0]),3);
    trh.in(3,&arr(v3f,ind[MESH_QUAD].data[2+c0]),3);
    trh.in(6,&arr(v3f,ind[MESH_QUAD].data[3+c0]),3);
    MESHGL_CENTROID
  }
  
  vol/=6;
  if (TENSOR_NO_DIVISION<abs(ret(3)))
    ret*=1/ret(3);
  else {
    ret=0;
    ret(3)=1;
    vol=0;
  }
  return ret;
}

tensor<double> meshgl::inertia(void) {
  tensor<int> idx=tensor<int>::index(1,ind[MESH_QUAD].size[1]);
  tensor<int> tri=ind[MESH_TRIA]&ind[MESH_QUAD](tensor<int>::vector(3).as(3,0,1),idx)&ind[MESH_QUAD](tensor<int>::vector(3).as(1,2,3),idx);
  int v3f=offsetV3F();
  tensor<double> ine=tensor<double>::matrix(3,3).as( 2.0, 1.0, 1.0,  1.0, 2.0, 1.0,  1.0, 1.0, 2.0);
  ine*=1.0/120;
  tensor<float> trh(2,3,3);
  tensor<double> trd;
  tensor<double> buf(2,3,3);
  double vol=0,det;
  for (int c0=0;c0<tri.size[1];++c0) {
    trh.in(0,&arr(v3f,tri(0,c0)),3);
    trh.in(3,&arr(v3f,tri(1,c0)),3);
    trh.in(6,&arr(v3f,tri(2,c0)),3);
    trd=trh.to<double>();
    det=trd.det();
    vol+=det;
    buf+=(trd*ine*trd.transpose())*det;
  }
  buf=tensor<double>::eye(3)*buf.trace(0,1).data[0]-buf;
  tensor<double> ret(2,4,4);
  ret.in(0,&buf.data[0],3);
  ret.in(4,&buf.data[3],3);
  ret.in(8,&buf.data[6],3);
  ret.data[15]=vol/6;
  return ret;
}

// more parts of it to tensor!
tensor<double> meshgl::bounding(void) {
  int v3f=offsetV3F();
  tensor<float> ret(2,4,2);
  if (arr.numel) { // TODO the empty mesh case is not canonic
    ret.in(0,arr^v3f,3);
    ret.in(4,arr^v3f,3);
    for (int c0=v3f;c0<arr.numel;c0+=arr.size[0]) {
      ret(0,0)=min(ret(0,0),arr.data[c0+0]);
      ret(0,1)=max(ret(0,1),arr.data[c0+0]);
      ret(1,0)=min(ret(1,0),arr.data[c0+1]);
      ret(1,1)=max(ret(1,1),arr.data[c0+1]);
      ret(2,0)=min(ret(2,0),arr.data[c0+2]);
      ret(2,1)=max(ret(2,1),arr.data[c0+2]);
    }
  }
  ret(3,0)=ret(3,1)=1;
  return ret.to<double>();
}

tensor<double> meshgl::vertices(void) {
  int v3f=offsetV3F();
  tensor<double> ret(2,4,arr.size[1]);
  for (int c0=0;c0<arr.size[1];++c0) {
    ret(0,c0)=arr(v3f+0,c0);
    ret(1,c0)=arr(v3f+1,c0);
    ret(2,c0)=arr(v3f+2,c0);
    ret(3,c0)=1;
  }
  return ret;
}

char* meshgl::toInfo(void) {
  sprintf(MESH_BUFFER,"%s V3F:%i N3F:%i",mesh<float>::toInfo(),offsetV3F(),offsetN3F());
  return MESH_BUFFER;
}

#include "meshgl_edit.hpp"
