// code by jph

#ifndef RANDOM_INCLUDED
#define RANDOM_INCLUDED

#define RANDOM_UNIFORM        314.15926535897932384

static char RANDOM_BUFFER[128];

template<class type>
class random {
  public:
    static type uniform(void);
    static type uniform(int bound);
    static type signum(void);
    static type normal(void);
    static char* toInfo(void); // ok
//----------------------------------------------------------------------
    static type seed;
  private:
    static int next;
    static type buffer;
};

template<class type>
type random<type>::uniform(void) { // returns numbers in the interval (0,1]
  if (RANDOM_UNIFORM<seed) // once for stability
    seed-=RANDOM_UNIFORM;  
  type ret=1000*sin(seed+=1);
  return 1-ret+floor(ret);
}

template<class type>
type random<type>::uniform(int bound) {
  return random<double>::uniform()*bound;
}

template<class type>
type random<type>::signum(void) {
  return 0.5<random<double>::uniform()?-1:1;
}

template<class type>
type random<type>::normal(void) {
  assert(sizeof(type)==8);
  next^=1;
  if (next) {
    type rad=sqrt(-2*log(uniform())); // 0 < uniform()  =>  log ok
    type ang=MATH_2PI*uniform();
    buffer=rad*cos(ang);
    return rad*sin(ang);
  } else
    return buffer;      
}

template<class type>
char* random<type>::toInfo(void) {
  sprintf(RANDOM_BUFFER,"random<%i>(%10.2e)",sizeof(type),seed);
  return RANDOM_BUFFER;
}

float random<float>::seed=time::now().s();
float random<float>::buffer=0;
int random<float>::next=0;

double random<double>::seed=time::now().s();
double random<double>::buffer=0;
int random<double>::next=0;

#endif
