// code by jph
// credits: shadow volumes after the paper by Cass Everitt and Mark J. Kildard,
//          Practical and Robust Stenciled Shadow Volumes for Hardware-Accelerated Rendering

// shadow volumes restricted to models with less than 65536 vertices
// rendering to the stencil buffer is slow on some gfx cards

class shadow : public meshgl {
public:
  shadow(meshgl* msh);
  void draw(tensor<double> rel);
  tensor<float> N3F; // normals per triangle
};

shadow::shadow(meshgl* msh) {
  assert(msh);
  tensor<int> idx=tensor<int>::index(1,msh->ind[MESH_QUAD].size[1]);
  ind[MESH_TRIA]=msh->ind[MESH_TRIA]&msh->ind[MESH_QUAD](tensor<int>::vector(3).as(3,0,1),idx)&msh->ind[MESH_QUAD](tensor<int>::vector(3).as(1,2,3),idx);
  arr=tensor<float>(2,4,msh->arr.size[1]);
  int v3f=msh->offsetV3F();
  for (int c0=0;c0<msh->arr.size[1];++c0) {
    arr(0,c0)=msh->arr(v3f+0,c0);
    arr(1,c0)=msh->arr(v3f+1,c0);
    arr(2,c0)=msh->arr(v3f+2,c0);
    arr(3,c0)=1;
  }  
  N3F=tensor<float>(2,4,ind[MESH_TRIA].size[1]);
  tensor<int> tri(1,3);
  for (int c0=0;c0<ind[MESH_TRIA].size[1];++c0) {
    tri.in(0,ind[MESH_TRIA]^c0*3,3);
    N3F.in(c0*4,(arr[tri(1)]-arr[tri(0)]).x(arr[tri(2)]-arr[tri(0)]).data,3);    
  }
}
                                                                                                                                             
#define MESHGL_SHADOW_TRIA              \
      t1=ind[MESH_TRIA].data[c0+1];     \
      t2=ind[MESH_TRIA].data[c0+2];     \
      v0.in(0,&arr.data[t0<<2],4);      \
      v1.in(0,&arr.data[t1<<2],4);      \
      v2.in(0,&arr.data[t2<<2],4);      \
      glVertex3fv(v0.data);             \
      glVertex3fv(v1.data);             \
      glVertex3fv(v2.data);             \
      v0-=pos;                          \
      glVertex4fv(v0.data);             \
      v2-=pos;                          \
      glVertex4fv(v2.data);             \
      v1-=pos;                          \
      glVertex4fv(v1.data);

#define MESHGL_SHADOW_EDGES             \
  glBegin(GL_QUADS);                    \
  for (int c0=0;c0<edg.numel;++c0) {    \
    t1=edg.data[c0]&0xffff;             \
    t0=edg.data[c0]>>16;                \
    vrt.in(0,&arr.data[t0<<2],4);       \
    glVertex3fv(vrt.data);              \
    vrt-=pos;                           \
    glVertex4fv(vrt.data);              \
    vrt.in(0,&arr.data[t1<<2],4);       \
    vrt-=pos;                           \
    glVertex4fv(vrt.data);              \
    vrt.in(0,&arr.data[t1<<2],3);       \
    glVertex3fv(vrt.data);              \
  }                                     \
  glEnd();

#define MESHGL_SHADOW_EDGE(X,Y)         \
      uori=X; uori<<=16; uori+=Y;       \
      if (X<Y)                          \
        eds.toggle(uori,uori);          \
      else {                            \
        usrt=Y; usrt<<=16; usrt+=X;     \
        eds.toggle(usrt,uori);          \
      }

int MESHGL_LIST[65536];

void shadow::draw(tensor<double> rel) {
  assert(arr.size[1]<0xffff);
  uint t0,t1,t2;
  hash<uint,uint> eds(sqrt((double)ind[MESH_TRIA].size[1])*7,set<uint>::map);
  tensor<float> pos(1,4);
  pos.in(0,rel.to<float>().data,3);
  pos(3)=1;
  tensor<float> vrt(1,4);
  uint uori,usrt;
  glCullFace(GL_FRONT);
  glStencilOp(GL_KEEP,GL_INCR,GL_KEEP);
  tensor<float> v0(1,4),v1(1,4),v2(1,4);
  int cnt=0;  
  //uint ret=glGenLists(1);
  //glNewList(ret,GL_COMPILE_AND_EXECUTE);  
  glBegin(GL_TRIANGLES);
  for (int c0=0;c0<ind[MESH_TRIA].numel;c0+=3) {
    t0=ind[MESH_TRIA].data[c0  ];
    vrt.in(0,&arr.data[t0<<2],4); 
    vrt-=pos;
    if ((N3F[c0/3]|vrt)<0) {
      MESHGL_LIST[cnt++]=c0;
      MESHGL_SHADOW_TRIA
      MESHGL_SHADOW_EDGE(t0,t1)
      MESHGL_SHADOW_EDGE(t1,t2)
      MESHGL_SHADOW_EDGE(t2,t0)
    }
  }
  glEnd();
  //printf("%s\n",eds.toInfo());
  tensor<uint> edg=eds.toTensor();
  MESHGL_SHADOW_EDGES  
  //glEndList();
  
  glCullFace(GL_BACK);
  glStencilOp(GL_KEEP,GL_DECR,GL_KEEP);
  //glCallList(ret);
  glBegin(GL_TRIANGLES);
  int c0;
  for (int c1=0;c1<cnt;++c1) {
    c0=MESHGL_LIST[c1];
    t0=ind[MESH_TRIA].data[c0];
    MESHGL_SHADOW_TRIA
  }
  glEnd();
  MESHGL_SHADOW_EDGES
  //glDeleteLists(ret,1);
}
